"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import grp
import os
import pathlib

import pytest
import pytest_inmanta.plugin

EXAMPLE_UNIT = """
[Unit]
Description=Podman inmanta-orchestrator-net.service
Documentation=https://github.com/edvgui/inmanta-module-podman
RequiresMountsFor=%t/containers
Wants=network-online.target
Wants=inmanta-orchestrator-db.service
Wants=inmanta-orchestrator-server.service
Before=inmanta-orchestrator-db.service
Before=inmanta-orchestrator-server.service
After=network-online.target

[Service]
Restart=on-failure
TimeoutStopSec=70
ExecStartPre=/usr/bin/podman network create --ignore --subnet=172.42.0.0/24 inmanta-orchestrator-net
ExecStart=/usr/bin/bash -c "/usr/bin/sleep infinity & /usr/bin/podman network inspect inmanta-orchestrator-net"
ExecStopPost=/usr/bin/podman network rm -f inmanta-orchestrator-net
Type=forking

[Install]
WantedBy=default.target

""".lstrip(
    "\n"
)


@pytest.mark.parametrize(
    (
        "file_path",
        "purged",
    ),
    [
        (pathlib.Path("/tmp/example"), False),
    ],
)
def test_model(
    project: pytest_inmanta.plugin.Project, file_path: pathlib.Path, purged: bool
) -> None:
    user = os.getlogin()
    group = grp.getgrgid(os.getgid()).gr_name
    model = f"""
        import mitogen
        import files
        import files::systemd_unit

        import std

        host = std::Host(
            name="localhost",
            os=std::linux,
            via=mitogen::Local(),
        )

        files::SystemdUnitFile(
            host=host,
            path={repr(str(file_path))},
            owner={repr(user)},
            group={repr(group)},
            purged={str(purged).lower()},
            unit=Unit(
                description="Podman inmanta-orchestrator-net.service",
                documentation=["https://github.com/edvgui/inmanta-module-podman"],
                requires_mounts_for=["%t/containers"],
                wants=[
                    "network-online.target",
                    "inmanta-orchestrator-db.service",
                    "inmanta-orchestrator-server.service",
                ],
                before=[
                    "inmanta-orchestrator-db.service",
                    "inmanta-orchestrator-server.service",
                ],
                after=["network-online.target"],
            ),
            service=Service(
                restart="on-failure",
                timeout_stop_sec=70,
                exec_start_pre=["/usr/bin/podman network create --ignore --subnet=172.42.0.0/24 inmanta-orchestrator-net"],
                exec_start="/usr/bin/bash -c \\"/usr/bin/sleep infinity & /usr/bin/podman network inspect inmanta-orchestrator-net\\"",
                exec_stop_post=["/usr/bin/podman network rm -f inmanta-orchestrator-net"],
                type="forking",
            ),
            install=Install(
                wanted_by=["default.target"],
            ),
        )
    """  # noqa: E501

    project.compile(model.strip("\n"), no_dedent=False)


def test_deploy(project: pytest_inmanta.plugin.Project, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "inmanta-orchestrator-net.service"

    # Create the file
    test_model(project, file, purged=False)
    assert project.dryrun_resource("files::SystemdUnitFile")
    project.deploy_resource("files::SystemdUnitFile")
    assert not project.dryrun_resource("files::SystemdUnitFile")

    # Check that the file content is the expected one
    assert file.read_text() == EXAMPLE_UNIT

    # Delete the file
    test_model(project, file, purged=True)
    assert project.dryrun_resource("files::SystemdUnitFile")
    project.deploy_resource("files::SystemdUnitFile")
    assert not project.dryrun_resource("files::SystemdUnitFile")
    assert not file.exists()



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import copy
import enum
import json
import typing

import inmanta_plugins.std
import yaml

import inmanta.agent.handler
import inmanta.execute.proxy
import inmanta.execute.util
import inmanta.plugins
import inmanta.resources
import inmanta_plugins.files.base
from inmanta.util import dict_path


@inmanta.plugins.plugin()
def get_json_fact(
    context: inmanta.plugins.Context,
    resource: typing.Annotated[typing.Any, inmanta.plugins.ModelType["std::Resource"]],
    fact_name: str,
    *,
    default_value: object | None = None,
    soft_fail: bool = False,
) -> object:
    """
    Get a value from fact that is expected to be a json-serialized payload.
    Deserialize the value and return it.
    If soft_fail is True and the value is not a valid json, return Unknown instead.

    :param resource: The resource that should provide the fact
    :param fact_name: The name of the fact provided by the resource
    :param default_value: A default value to return if the fact is not set yet
    :param soft_fail: Whether to suppress json decoding error and return Unknown instead.
    """
    # Get the fact using std logic
    fact = inmanta_plugins.std.getfact(
        context,
        resource,
        fact_name,
    )

    # If the fact is unknown and we have a default, we return the default
    # instead
    if inmanta_plugins.std.is_unknown(fact) and default_value is not None:
        return default_value

    # If the fact is unknown, we return it as is
    if inmanta_plugins.std.is_unknown(fact):
        return fact

    # Try to decode the json
    try:
        return json.loads(fact)
    except json.JSONDecodeError:
        if soft_fail:
            # Return unknown instead
            return inmanta.execute.util.Unknown(source=resource)

        raise


class Operation(str, enum.Enum):
    REPLACE = "replace"
    REMOVE = "remove"
    MERGE = "merge"


def update(
    config: dict, path: dict_path.DictPath, operation: Operation, desired: object
) -> dict:
    """
    Update the config config at the specified type, using given operation and desired value.

    :param config: The configuration to update
    :param path: The path pointing to an element of the config that should be modified
    :param operation: The type of operation to apply to the config element
    :param desired: The desired state to apply to the config element
    """
    if operation == Operation.REMOVE:
        path.remove(config)
        return config

    if operation == Operation.REPLACE:
        path.set_element(config, value=desired)
        return config

    if operation == Operation.MERGE:
        if not isinstance(desired, dict):
            raise ValueError(
                f"Merge operation is only supported for dicts, but got {type(desired)} "
                f"({desired})"
            )
        current = path.get_element(config, construct=True)
        if not isinstance(current, dict):
            raise ValueError(
                f"A dict can only me merged to a dict, current value at path {path} "
                f"is not a dict: {current} ({type(current)})"
            )
        current.update({k: v for k, v in desired.items() if v is not None})
        return config

    raise ValueError(f"Unsupported operation: {operation}")


@inmanta.resources.resource(
    name="files::JsonFile",
    id_attribute="path",
    agent="host.name",
)
class JsonFileResource(inmanta_plugins.files.base.BaseFileResource):
    fields = (
        "indent",
        "format",
        "values",
        "discovered_values",
    )
    values: list[dict]
    discovered_values: list[dict]
    format: typing.Literal["json", "yaml"]
    indent: int

    @classmethod
    def get_values(cls, _, entity: inmanta.execute.proxy.DynamicProxy) -> list[dict]:
        return [
            {
                "path": value.path,
                "operation": value.operation,
                "value": value.value,
            }
            for value in entity.values
        ]

    @classmethod
    def get_discovered_values(
        cls, _, entity: inmanta.execute.proxy.DynamicProxy
    ) -> list[dict]:
        return [
            {
                "path": value.path,
            }
            for value in entity.discovered_values
        ]


@inmanta.resources.resource(
    name="files::SharedJsonFile",
    id_attribute="uri",
    agent="host.name",
)
class SharedJsonFileResource(JsonFileResource):
    fields = ("uri",)

    @classmethod
    def get_uri(cls, _, entity: inmanta.execute.proxy.DynamicProxy) -> str:
        """
        Compose a uri to identify the resource, and which allows multiple resources
        to manage the same file.
        """
        if entity.resource_discriminator:
            return f"{entity.path}:{entity.resource_discriminator}"
        return entity.path


@inmanta.agent.handler.provider("files::JsonFile", "")
@inmanta.agent.handler.provider("files::SharedJsonFile", "")
class JsonFileHandler(inmanta_plugins.files.base.BaseFileHandler[JsonFileResource]):
    def from_json(self, raw: str, *, format: typing.Literal["json", "yaml"]) -> object:
        """
        Convert a json-like raw string in the expected format to the corresponding
        python dict-like object.

        :param raw: The raw value, as read in the file.
        :param format: The format of the value.
        """
        if format == "json":
            return json.loads(raw)
        if format == "yaml":
            return yaml.safe_load(raw)
        raise ValueError(f"Unsupported format: {format}")

    def to_json(
        self,
        value: object,
        *,
        format: typing.Literal["json", "yaml"],
        indent: typing.Optional[int] = None,
    ) -> str:
        """
        Dump a dict-like structure into a json-like string.  The string can
        be in different formats, depending on the value specified.

        :param value: The dict-like value, to be written to file.
        :param format: The format of the value.
        :param indent: Whether any indentation should be applied to the
            value written to file.
        """
        if format == "json":
            return json.dumps(value, indent=indent)
        if format == "yaml":
            return yaml.safe_dump(value, indent=indent)
        raise ValueError(f"Unsupported format: {format}")

    def extract_facts(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: JsonFileResource,
        *,
        content: dict,
    ) -> dict[str, str]:
        # Read facts based on the content of the file
        return {
            str(path): json.dumps(
                {
                    str(k): dict_path.to_path(str(k)).get_element(content)
                    for k in path.resolve_wild_cards(content)
                }
            )
            for desired_value in resource.discovered_values
            if (path := dict_path.to_wild_path(desired_value["path"]))
        }

    def facts(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: JsonFileResource,
    ) -> dict[str, object]:
        try:
            # Delegate to read_resource to get the content of
            # the file
            self.read_resource(ctx, resource)
        except inmanta.agent.handler.ResourcePurged():
            return {}

        return self.extract_facts(
            ctx,
            resource,
            content=ctx.get("current_content"),
        )

    def read_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: JsonFileResource,
    ) -> None:
        super().read_resource(ctx, resource)

        # Load the content of the existing file
        raw_content = self.proxy.read_binary(resource.path).decode()
        ctx.debug("Reading existing file", raw_content=raw_content)
        current_content = self.from_json(raw_content, format=resource.format)
        ctx.set("current_content", current_content)

    def calculate_diff(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        current: JsonFileResource,
        desired: JsonFileResource,
    ) -> dict[str, dict[str, object]]:
        # For file permissions and ownership, we delegate to the parent class
        changes = super().calculate_diff(ctx, current, desired)

        # To check if some change content needs to be applied, we perform a "stable" addition
        # operation: We apply our desired state to the current state, and check if we can then
        # see any difference.
        current_content = ctx.get("current_content")
        desired_content = copy.deepcopy(current_content)
        for value in desired.values:
            update(
                desired_content,
                dict_path.to_path(value["path"]),
                Operation(value["operation"]),
                value["value"],
            )

        if current_content != desired_content:
            changes["content"] = {
                "current": current_content,
                "desired": desired_content,
            }

        # Set the facts now if it is a dryrun or if there is
        # no changes
        if not changes or ctx.is_dry_run():
            for k, v in self.extract_facts(
                ctx,
                desired,
                content=current_content,
            ).items():
                ctx.set_fact(k, v)

        return changes

    def create_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: JsonFileResource,
    ) -> None:
        # Build a config based on all the elements we want to manage
        content = {}
        for value in resource.values:
            update(
                content,
                dict_path.to_path(value["path"]),
                Operation(value["operation"]),
                value["value"],
            )

        indent = resource.indent if resource.indent != 0 else None
        raw_content = self.to_json(
            content,
            format=resource.format,
            indent=indent,
        )
        self.proxy.put(resource.path, raw_content.encode())
        super().create_resource(ctx, resource)

        # Set the facts after creation
        for k, v in self.extract_facts(ctx, resource, content=content).items():
            ctx.set_fact(k, v)

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: JsonFileResource,
    ) -> None:
        if "content" in changes:
            content = changes["content"]["desired"]
            indent = resource.indent if resource.indent != 0 else None
            raw_content = self.to_json(
                content,
                format=resource.format,
                indent=indent,
            )
            self.proxy.put(resource.path, raw_content.encode())

            # Set the facts after update
            for k, v in self.extract_facts(ctx, resource, content=content).items():
                ctx.set_fact(k, v)

        super().update_resource(ctx, changes, resource)



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import typing

import inmanta_plugins.mitogen.abc

import inmanta.agent.agent
import inmanta.agent.handler
import inmanta.const
import inmanta.execute.proxy
import inmanta.export
import inmanta.resources


class BaseFileResource(
    inmanta_plugins.mitogen.abc.ResourceABC, inmanta.resources.ManagedResource
):
    fields = ("path", "permissions", "owner", "group")
    path: str
    permissions: typing.Optional[int]
    owner: typing.Optional[str]
    group: typing.Optional[str]


X = typing.TypeVar("X", bound=BaseFileResource)


class BaseFileHandler(inmanta_plugins.mitogen.abc.HandlerABC[X]):
    def whoami(self) -> str:
        """
        Check which user is currently running the the commands on the proxy.
        The result is cached on the proxy object to avoid running the command
        more times than required.
        """
        if not hasattr(self.proxy, "_whoami"):
            stdout, stderr, code = self.proxy.run("whoami")
            if code != 0:
                raise RuntimeError(
                    f"Failed to check current user on the remote host: {stderr}"
                )

            # Cache the result
            setattr(self.proxy, "_whoami", stdout)

        return typing.cast(str, getattr(self.proxy, "_whoami"))

    def read_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: X
    ) -> None:
        if not self.proxy.file_exists(resource.path):
            raise inmanta.agent.handler.ResourcePurged()

        for key, value in self.proxy.file_stat(resource.path).items():
            if getattr(resource, key) is not None:
                setattr(resource, key, value)

    def create_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: X
    ) -> None:
        if resource.permissions is not None:
            self.proxy.chmod(resource.path, str(resource.permissions))

        if resource.owner is not None or resource.group is not None:
            self.proxy.chown(resource.path, resource.owner, resource.group)

        ctx.set_created()

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: X,
    ) -> None:
        if "permissions" in changes:
            self.proxy.chmod(resource.path, str(resource.permissions))

        if "owner" in changes or "group" in changes:
            self.proxy.chown(resource.path, resource.owner, resource.group)

        ctx.set_updated()

    def delete_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: X
    ) -> None:
        self.proxy.remove(resource.path)
        ctx.set_purged()



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import grp
import os
import pathlib
import shutil

import pytest_inmanta.plugin


def test_model(
    project: pytest_inmanta.plugin.Project,
    dir_path: pathlib.Path = pathlib.Path("/tmp/example"),
    create_parents: bool = False,
    purged: bool = False,
) -> None:
    user = os.getlogin()
    group = grp.getgrgid(os.getgid()).gr_name
    model = f"""
        import mitogen
        import files

        import std

        host = std::Host(
            name="localhost",
            os=std::linux,
            via=mitogen::Local(),
        )

        files::Directory(
            host=host,
            path={repr(str(dir_path))},
            owner={repr(user)},
            group={repr(group)},
            purged={str(purged).lower()},
            create_parents={str(create_parents).lower()},
        )
    """

    project.compile(model.strip("\n"), no_dedent=False)


def test_deploy(project: pytest_inmanta.plugin.Project, tmp_path: pathlib.Path) -> None:
    dir = tmp_path / "dir"

    # Create the dir
    test_model(project, dir, purged=False, create_parents=False)
    assert project.dryrun_resource("files::Directory")
    project.deploy_resource("files::Directory")
    assert dir.is_dir()
    assert not project.dryrun_resource("files::Directory")

    # Create the dir (recursively)
    test_model(project, dir / "in/dir", purged=False, create_parents=True)
    assert not (dir / "in").exists()
    assert project.dryrun_resource("files::Directory")
    project.deploy_resource("files::Directory")
    assert (dir / "in/dir").is_dir()
    assert (dir / "in").owner() == dir.owner()
    assert (dir / "in").group() == dir.group()
    assert (dir / "in").stat().st_mode == dir.stat().st_mode
    assert not project.dryrun_resource("files::Directory")

    # Create another directory recursively in a folder that is owned
    # by root, but running the test as a non-root user
    other_tmp_path = pathlib.Path("/tmp/test")
    if other_tmp_path.exists():
        shutil.rmtree(str(other_tmp_path))

    test_model(project, other_tmp_path / "a", create_parents=True)
    project.deploy_resource("files::Directory")

    # Delete the dir
    test_model(project, dir, purged=True)
    assert project.dryrun_resource("files::Directory")
    project.deploy_resource("files::Directory")
    assert not dir.exists()
    assert not project.dryrun_resource("files::Directory")



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import grp
import os
import pathlib

import pytest
import pytest_inmanta.plugin


@pytest.mark.parametrize(
    (
        "file_path",
        "purged",
    ),
    [
        (pathlib.Path("/tmp/example"), False),
    ],
)
def test_model(
    project: pytest_inmanta.plugin.Project, file_path: pathlib.Path, purged: bool
) -> None:
    user = os.getlogin()
    group = grp.getgrgid(os.getgid()).gr_name
    model = f"""
        import mitogen
        import files
        import files::host

        import std

        host = std::Host(
            name="localhost",
            os=std::linux,
            via=mitogen::Local(),
        )

        files::SharedHostFile(
            host=host,
            path={repr(str(file_path))},
            owner={repr(user)},
            group={repr(group)},
            purged={str(purged).lower()},
            entries=[
                files::host::Entry(
                    hostname="example.com",
                    address4="192.168.10.10",
                    operation="replace",
                ),
            ],
        )

        files::SharedHostFile(
            host=host,
            path={repr(str(file_path))},
            resource_discriminator="example.be",
            purged={str(purged).lower()},
            entries=[
                files::host::Entry(
                    hostname="example.be",
                    operation="remove",
                ),
            ],
        )

        files::SharedHostFile(
            host=host,
            path={repr(str(file_path))},
            resource_discriminator="example.eu",
            purged={str(purged).lower()},
            entries=[
                files::host::Entry(
                    hostname="example.eu",
                    address4="192.168.10.10",
                    operation="merge",
                ),
            ],
        )
    """

    project.compile(model.strip("\n"), no_dedent=False)


def test_deploy(project: pytest_inmanta.plugin.Project, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "host"

    # Create the file
    test_model(project, file, purged=False)
    assert project.dryrun_resource("files::SharedHostFile", uri=str(file))
    assert project.dryrun_resource("files::SharedHostFile", uri=f"{file}:example.be")
    assert project.dryrun_resource("files::SharedHostFile", uri=f"{file}:example.eu")
    project.deploy_resource("files::SharedHostFile", uri=str(file))
    project.deploy_resource("files::SharedHostFile", uri=f"{file}:example.be")
    project.deploy_resource("files::SharedHostFile", uri=f"{file}:example.eu")
    assert not project.dryrun_resource("files::SharedHostFile", uri=str(file))
    assert not project.dryrun_resource(
        "files::SharedHostFile", uri=f"{file}:example.be"
    )
    assert not project.dryrun_resource(
        "files::SharedHostFile", uri=f"{file}:example.eu"
    )



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import pathlib

import inmanta.plugins


@inmanta.plugins.plugin()
def path_join(base_path: str, *extra: str) -> str:
    """
    Join together the base_path and all of the extra parts after it.  If any extra
    item specified an absolute path (starts with a '/') it will overwrite all the
    elements of the path before it.

    :param base_path: The base path, a directory, to which should be
        appended all the extra items.
    :param *extra: A set of extra parts to add to the path.
    """
    return str(pathlib.Path(base_path, *extra))



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import grp
import json
import os
import pathlib

import pytest_inmanta.plugin


def test_model(
    project: pytest_inmanta.plugin.Project,
    file_path: pathlib.Path = pathlib.Path("/tmp/example"),
    purged: bool = False,
    format: str = "json",
    people_fact: str | None = None,
) -> None:
    user = os.getlogin()
    group = grp.getgrgid(os.getgid()).gr_name
    model = f"""
        import mitogen
        import files
        import files::json

        import std

        host = std::Host(
            name="localhost",
            os=std::linux,
            via=mitogen::Local(),
        )

        files::JsonFile(
            host=host,
            path={repr(str(file_path))},
            owner={repr(user)},
            group={repr(group)},
            purged={str(purged).lower()},
            format={repr(format)},
            values=[
                files::json::Object(
                    path="people[name=bob]",
                    operation="replace",
                    value={{"name": "bob", "age": 20}},
                ),
                files::json::Object(
                    path="people[name=alice]",
                    operation="merge",
                    value={{"name": "alice", "age": 20}},
                ),
                files::json::Object(
                    path="people[name=eve]",
                    operation="remove",
                    value={{}},
                ),
            ],
            discovered_values=[
                files::json::DiscoveredValue(
                    path="people[name=*]",
                ),
            ],
        )
    """

    if people_fact is not None:
        # If the fact is set for the people, we add it to the project
        project.add_fact(
            f"files::JsonFile[localhost,path={file_path}]",
            "people[name=*]",
            people_fact,
        )

    project.compile(model.strip("\n"), no_dedent=False)

    if people_fact is not None:
        # Check that the fact has been assigned in the model
        discovered = project.get_instances("files::json::DiscoveredValue")
        assert len(discovered) == 1
        assert dict(discovered[0].values) == json.loads(people_fact)


def test_deploy(project: pytest_inmanta.plugin.Project, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "friends.json"
    res_id = f"files::JsonFile[localhost,path={file}]"

    def get_people_fact_value(resource_id: str) -> dict | None:
        if project.ctx is None:
            # No handler has run, the fact can't be set
            return None

        for fact in project.ctx.facts:
            if fact["resource_id"] != resource_id:
                continue

            if fact["id"] != "people[name=*]":
                continue

            return json.loads(fact["value"])

        # No fact matched
        return None

    # Create the file
    test_model(project, file, purged=False)
    assert project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) is None
    project.deploy_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
    }
    assert not project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
    }

    # Manually remove a managed line from the file and make sure we detect a change
    friends = json.loads(file.read_text())
    del friends["people"]
    file.write_text(json.dumps(friends))
    assert project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {}
    project.deploy_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
    }
    assert not project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
    }

    # Insert an extra entry in the file and me sure we don't detect any change as we don't
    # manage that entry
    friends = json.loads(file.read_text())
    friends["people"].append({"name": "chris"})
    file.write_text(json.dumps(friends))
    assert not project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
        "people[name=chris]": {"name": "chris"},
    }

    # Add the entry that should not be there and make sure it is removed, the unmanaged
    # entry should remain untouched
    friends["people"].append({"name": "eve"})
    file.write_text(json.dumps(friends))
    assert project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
        "people[name=chris]": {"name": "chris"},
        "people[name=eve]": {"name": "eve"},
    }
    project.deploy_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
        "people[name=chris]": {"name": "chris"},
    }
    assert not project.dryrun_resource("files::JsonFile")
    friends = json.loads(file.read_text())
    assert friends["people"][2] == {"name": "chris"}
    assert len(friends["people"]) == 3

    # Delete the file
    test_model(
        project,
        file,
        purged=True,
        people_fact=json.dumps(
            {
                "people[name=bob]": {"name": "bob", "age": 20},
                "people[name=alice]": {"name": "alice", "age": 20},
                "people[name=chris]": {"name": "chris"},
            },
        ),
    )
    assert project.dryrun_resource("files::JsonFile")
    assert get_people_fact_value(res_id) == {
        "people[name=bob]": {"name": "bob", "age": 20},
        "people[name=alice]": {"name": "alice", "age": 20},
        "people[name=chris]": {"name": "chris"},
    }
    project.deploy_resource("files::JsonFile")
    assert get_people_fact_value(res_id) is None
    assert not project.dryrun_resource("files::JsonFile")
    assert not file.exists()

    # Create the file as a yaml
    test_model(project, file, purged=False, format="yaml")
    assert project.dryrun_resource("files::JsonFile")
    project.deploy_resource("files::JsonFile")
    assert not project.dryrun_resource("files::JsonFile")



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import pathlib

import inmanta.agent.agent
import inmanta.agent.handler
import inmanta.const
import inmanta.execute.proxy
import inmanta.export
import inmanta.resources
import inmanta_plugins.files.base
import inmanta_plugins.files.json


@inmanta.resources.resource(
    name="files::Directory",
    id_attribute="path",
    agent="host.name",
)
class DirectoryResource(inmanta_plugins.files.base.BaseFileResource):
    fields = ("create_parents",)
    create_parents: bool


@inmanta.agent.handler.provider("files::Directory", "")
class DirectoryHandler(inmanta_plugins.files.base.BaseFileHandler[DirectoryResource]):
    def read_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: DirectoryResource
    ) -> None:
        super().read_resource(ctx, resource)

    def create_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: DirectoryResource
    ) -> None:
        if resource.create_parents:
            # Create all parent directories, with the owner, group and permissions of
            # their parents
            parent_owner = "root"
            parent_group = "root"
            parent_permissions = "555"
            for parent in reversed(pathlib.Path(resource.path).parents):
                if not self.proxy.file_exists(str(parent)):
                    # Create the parent directory, and make sure it has the
                    # right owner and permissions
                    self.proxy.mkdir(str(parent))

                    if self.whoami() == "root":
                        # We can only change the ownership if we are root
                        self.proxy.chown(str(parent), parent_owner, parent_group)

                    self.proxy.chmod(str(parent), parent_permissions)
                    continue

                # Read the existing folder permissions, and save it for the next child folder
                stat = self.proxy.file_stat(str(parent))
                parent_owner = stat["owner"]
                parent_group = stat["group"]
                parent_permissions = str(stat["permissions"])

        # Call the basic io mkdir helper
        self.proxy.mkdir(resource.path)

        super().create_resource(ctx, resource)

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: DirectoryResource,
    ) -> None:
        super().update_resource(ctx, changes, resource)

    def delete_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: DirectoryResource
    ) -> None:
        self.proxy.rmdir(resource.path)
        ctx.set_purged()



"""
Copyright 2024 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import inmanta.agent.agent
import inmanta.agent.handler
import inmanta.const
import inmanta.execute.proxy
import inmanta.export
import inmanta.resources
import inmanta_plugins.files.base


@inmanta.resources.resource(
    name="files::SystemdUnitFile",
    id_attribute="path",
    agent="host.name",
)
class SystemdUnitFileResource(inmanta_plugins.files.base.BaseFileResource):
    fields = ("content",)
    content: str


@inmanta.agent.handler.provider("files::SystemdUnitFile", "")
class SystemdUnitFileHandler(
    inmanta_plugins.files.base.BaseFileHandler[SystemdUnitFileResource]
):
    def read_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: SystemdUnitFileResource,
    ) -> None:
        super().read_resource(ctx, resource)

        # Load the content of the existing file
        resource.content = self.proxy.read_binary(resource.path).decode()
        ctx.debug("Reading existing file", content=resource.content)

    def create_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        resource: SystemdUnitFileResource,
    ) -> None:
        self.proxy.put(resource.path, resource.content.encode())
        super().create_resource(ctx, resource)

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: SystemdUnitFileResource,
    ) -> None:
        if "content" in changes:
            self.proxy.put(resource.path, resource.content.encode())

        super().update_resource(ctx, changes, resource)



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import inmanta.agent.agent
import inmanta.agent.handler
import inmanta.const
import inmanta.execute.proxy
import inmanta.export
import inmanta.resources
import inmanta_files
import inmanta_plugins.files.base
import inmanta_plugins.files.json


@inmanta.resources.resource(
    name="files::Symlink",
    id_attribute="path",
    agent="host.name",
)
class SymlinkResource(inmanta_plugins.files.base.BaseFileResource):
    fields = ("target",)
    target: str


@inmanta.agent.handler.provider("files::Symlink", "")
class SymlinkHandler(inmanta_plugins.files.base.BaseFileHandler[SymlinkResource]):
    def chown_symlink(self, path: str, owner: str, group: str) -> None:
        """
        Perform the chown operation on a symlink, it requires an extra argument
        to prevent chown from resolving the symlink itself and changing the
        permissions of the target file.
        """
        self.proxy.run(
            "chown",
            [
                "--no-dereference",
                f"{owner}:{group}",
                path,
            ],
        )

    def read_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: SymlinkResource
    ) -> None:
        if not self.proxy.file_exists(resource.path):
            raise inmanta.agent.handler.ResourcePurged()

        if not self.proxy.is_symlink(resource.path):
            raise Exception(
                "The target of resource %s already exists but is not a symlink."
                % resource
            )

        for key, value in self.proxy.remote_call(
            inmanta_files.symlink_stat,
            resource.path,
        ).items():
            if getattr(resource, key) is not None:
                setattr(resource, key, value)

        resource.target = self.proxy.readlink(resource.path)

    def create_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: SymlinkResource
    ) -> None:
        # Call the basic io symlink helper
        self.proxy.symlink(resource.target, resource.path)

        if resource.owner is not None or resource.group is not None:
            self.chown_symlink(resource.path, resource.owner, resource.group)

        ctx.set_created()

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: SymlinkResource,
    ) -> None:
        if "target" in changes:
            self.proxy.remove(resource.path)
            self.proxy.symlink(resource.target, resource.path)

        if "owner" in changes or "group" in changes:
            self.chown_symlink(resource.path, resource.owner, resource.group)

        ctx.set_updated()

    def delete_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: SymlinkResource
    ) -> None:
        self.proxy.remove(resource.path)
        ctx.set_purged()



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com

This module contains some code that will be executed on a remote host
using mitogen.
"""

import grp
import os
import pwd


def symlink_stat(path: str) -> dict[str, object]:
    """
    This method is similar to inmanta_mitogen.file_stat excepts that it doesn't
    try to resolve the symlink on the last element of the path (all other
    symlinks will be resolved).

    :param path: The path to stat.
    """
    # Resolve all the symlinks except the one at the end of the path
    parent_path = os.path.abspath(os.path.join(path, os.pardir))
    resolved_parent_path = os.path.realpath(parent_path)

    # Get the stat result for the symlink at the end of the path
    stat_result = os.stat(
        os.path.join(resolved_parent_path, os.path.basename(path)),
        follow_symlinks=False,
    )
    return dict(
        owner=pwd.getpwuid(stat_result.st_uid).pw_name,
        group=grp.getgrgid(stat_result.st_gid).gr_name,
        permissions=int(oct(stat_result.st_mode)[-4:]),
    )



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

from pytest_inmanta.plugin import Project

import inmanta_plugins.files


def test_basics(project: Project) -> None:
    project.compile("import files")


def test_path_join() -> None:
    assert inmanta_plugins.files.path_join("/test", "a", "b") == "/test/a/b"
    assert inmanta_plugins.files.path_join("/test", "/a", "b") == "/a/b"
    assert inmanta_plugins.files.path_join("test", "a", "b") == "test/a/b"
    assert inmanta_plugins.files.path_join("test/", "a", "b") == "test/a/b"



"""
Copyright 2023 Guillaume Everarts de Velp

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: edvgui@gmail.com
"""

import collections
import copy
import ipaddress
import typing

import inmanta.agent.agent
import inmanta.agent.handler
import inmanta.const
import inmanta.execute.proxy
import inmanta.export
import inmanta.resources
import inmanta_plugins.files.base
import inmanta_plugins.files.json
from inmanta.util import dict_path

PARSED_HOST_FILE = dict[str, dict[str, typing.Optional[str]]]
"""
Define an alias for the parsed host file format.  The parsed host file
is a dict which has as keys, hostnames, and as values, a dict containing
two keys: address4 and address6, respectively representing the ipv4 and
ipv6 address defined in the file for the given hostname.

And example of such dict is:

.. code-block::python

    {
        "localhost": {"address4": "127.0.0.1", "address6": "::1"},
        "ip6-loopback": {"address4": None, "address6": "::1"},
        "example.com": {"address4": "93.184.216.34", "address6": None},
    }

"""


def parse_host_file(raw_content: str) -> PARSED_HOST_FILE:
    """
    Parse the content of an host file, and return it as a dict having as
    keys the hostnames, and as values, the corresponding ips.
    """
    parsed_file: PARSED_HOST_FILE = collections.defaultdict(
        lambda: {
            "address4": None,
            "address6": None,
        }
    )

    for line in raw_content.splitlines():
        if not line:
            # No info in that line
            continue

        if line[0] not in "0123456789:":
            # This is not a valid entry, probably a comment
            continue

        parsed_line = line.split()

        # Parse the ip address
        address = ipaddress.ip_address(parsed_line[0])

        # For each hostname, add the ipaddress to the parsed file
        for hostname in parsed_line[1:]:
            parsed_file[hostname][f"address{address.version}"] = str(address)

    return dict(parsed_file)


def write_host_file(parsed_file: PARSED_HOST_FILE) -> str:
    """
    Generate the content of the hostfile in a single string, that
    can be written down into a file.
    """
    raw_content = ""

    for hostname, addresses in parsed_file.items():
        if addresses.get("address4") is not None:
            raw_content += f"{addresses['address4']} {hostname}\n"

        if addresses.get("address6") is not None:
            raw_content += f"{addresses['address6']} {hostname}\n"

    return raw_content


@inmanta.resources.resource(
    name="files::HostFile",
    id_attribute="path",
    agent="host.name",
)
class HostFileResource(inmanta_plugins.files.base.BaseFileResource):
    fields = ("values",)
    values: list[dict]

    @classmethod
    def get_values(
        cls,
        _: inmanta.export.Exporter,
        entity: inmanta.execute.proxy.DynamicProxy,
    ) -> list[dict]:
        return [
            {
                "path": str(dict_path.InDict(entry.hostname)),
                "operation": entry.operation,
                "value": {
                    "address4": entry.address4,
                    "address6": entry.address6,
                },
            }
            for entry in entity.entries
        ]


@inmanta.resources.resource(
    name="files::SharedHostFile",
    id_attribute="uri",
    agent="host.name",
)
class SharedHostFileResource(HostFileResource):
    fields = ("uri",)
    uri: str

    @classmethod
    def get_uri(cls, _, entity: inmanta.execute.proxy.DynamicProxy) -> str:
        """
        Compose a uri to identify the resource, and which allows multiple resources
        to manage the same file.
        """
        if entity.resource_discriminator:
            return f"{entity.path}:{entity.resource_discriminator}"
        return entity.path


@inmanta.agent.handler.provider("files::HostFile", "")
@inmanta.agent.handler.provider("files::SharedHostFile", "")
class HostFileHandler(inmanta_plugins.files.base.BaseFileHandler[HostFileResource]):
    def read_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: HostFileResource
    ) -> None:
        super().read_resource(ctx, resource)

        # Load the content of the existing file
        raw_content = self.proxy.read_binary(resource.path).decode()
        ctx.debug("Reading existing file", raw_content=raw_content)
        ctx.set("current_content", parse_host_file(raw_content))

    def calculate_diff(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        current: HostFileResource,
        desired: HostFileResource,
    ) -> dict[str, dict[str, object]]:
        # For file permissions and ownership, we delegate to the parent class
        changes = super().calculate_diff(ctx, current, desired)

        # To check if some change content needs to be applied, we perform a "stable" addition
        # operation: We apply our desired state to the current state, and check if we can then
        # see any difference.
        current_content = ctx.get("current_content")
        desired_content = copy.deepcopy(current_content)

        for value in desired.values:
            inmanta_plugins.files.json.update(
                desired_content,
                dict_path.to_path(value["path"]),
                inmanta_plugins.files.json.Operation(value["operation"]),
                value["value"],
            )

        if current_content != desired_content:
            changes["content"] = {
                "current": current_content,
                "desired": desired_content,
            }

        return changes

    def create_resource(
        self, ctx: inmanta.agent.handler.HandlerContext, resource: HostFileResource
    ) -> None:
        # Build a config based on all the values we want to manage
        content = {}
        for value in resource.values:
            inmanta_plugins.files.json.update(
                content,
                dict_path.to_path(value["path"]),
                inmanta_plugins.files.json.Operation(value["operation"]),
                value["value"],
            )
        raw_content = write_host_file(content)
        self.proxy.put(resource.path, raw_content.encode())
        super().create_resource(ctx, resource)

    def update_resource(
        self,
        ctx: inmanta.agent.handler.HandlerContext,
        changes: dict[str, dict[str, object]],
        resource: HostFileResource,
    ) -> None:
        if "content" in changes:
            raw_content = write_host_file(changes["content"]["desired"])
            self.proxy.put(resource.path, raw_content.encode())

        super().update_resource(ctx, changes, resource)
